/****************************************************************************

 Copyright 2003 Matthew Riek
 See legal notice in copyright.txt

 Description : micro xml dom

 ****************************************************************************/

#include "mxmldom.h"
#include <string.h>
#include <new.h>
#include <assert.h>
#include <stdarg.h>
#include <stdio.h>

void MXMLNode::Dispose()
{
  if(m_parent)
  {
    m_parent->RemoveChild(this);
  }
  MXMLDocument * document = m_document;
  this->~MXMLNode();
  document->Free(this);
}


MXMLNode::MXMLNode(MXMLDocument * a_document)
{
  m_type = T_NODE;
  m_document = a_document;
  m_parent = NULL;
  m_nextSibling = NULL;
  m_prevSibling = NULL;
}


void MXMLTextBase::SetText(const char * a_text, size_t a_size)
{
  const char * text = m_text;
  m_text = NULL;
  if(a_text)
  {
    size_t size = (a_size) ? a_size : MXMLUtf8Size(a_text);
    char * text = (char *) m_document->Alloc(size + 1);
    memcpy(text, a_text, size);
    text[size] = '\0';
    m_text = text;
  }
  if(text)
  {
    m_document->Free((void *) text);
  }
}


MXMLTextBase::MXMLTextBase(MXMLDocument * a_document) : MXMLNode(a_document)
{
  m_text = NULL;
}


MXMLTextBase::~MXMLTextBase()
{
  if(m_text)
  {
    m_document->Free((void *) m_text);
  }
}


bool MXMLDeclaration::Write(MXMLStream &a_stream) const
{
  int result = 0;
  size_t size = 0;
  const char * open = "<?";
  const char * close = "?>";
  const char * text = GetText();

  size = strlen(open);
  result |= (a_stream.Write(open, size) != size);

  size = MXMLUtf8Size(text);
  result |= (a_stream.Write(text, size) != size);

  size = strlen(close);
  result |= (a_stream.Write(close, size) != size);

  return (result == 0);
}


MXMLDeclaration::MXMLDeclaration(MXMLDocument * a_document) : MXMLTextBase(a_document)
{
  m_type |= T_DECLARATION;
}


bool MXMLComment::Write(MXMLStream &a_stream) const
{
  int result = 0;
  size_t size = 0;
  const char * open = "<!--";
  const char * close = "-->";
  const char * text = GetText();

  size = strlen(open);
  result |= (a_stream.Write(open, size) != size);

  size = MXMLUtf8Size(text);
  result |= (a_stream.Write(text, size) != size);

  size = strlen(close);
  result |= (a_stream.Write(close, size) != size);

  return (result == 0);
}


MXMLComment::MXMLComment(MXMLDocument * a_document) : MXMLTextBase(a_document)
{
  m_type |= T_COMMENT;
}


bool MXMLText::Write(MXMLStream &a_stream) const
{
  int result = 0;
  size_t size = 0;
  const char * text = GetText();

  size = MXMLUtf8Size(text);
  result |= (a_stream.Write(text, size) != size);

  return (result == 0);
}


MXMLText::MXMLText(MXMLDocument * a_document) : MXMLTextBase(a_document)
{
  m_type |= T_TEXT;
}


bool MXMLCDATA::Write(MXMLStream &a_stream) const
{
  int result = 0;
  size_t size = 0;
  const char * open = "<![CDATA[";
  const char * close = "]]>";
  const char * text = GetText();

  size = strlen(open);
  result |= (a_stream.Write(open, size) != size);

  size = MXMLUtf8Size(text);
  result |= (a_stream.Write(text, size) != size);

  size = strlen(close);
  result |= (a_stream.Write(close, size) != size);

  return (result == 0);
}


MXMLCDATA::MXMLCDATA(MXMLDocument * a_document) : MXMLText(a_document)
{
  m_type |= T_CDATA;
}


const char * MXMLElement::Attributes::GetNamedItem(const char * a_name, const char * a_default) const
{
  Attribute * attribute = m_attributes;
  while(attribute)
  {
    if(strcmp(attribute->GetName(), a_name) == 0)
    {
      if(attribute->m_value)
      {
        return attribute->m_value;
      }
      break;
    }
    attribute = attribute->m_next;
  }
  return a_default;
}


void MXMLElement::Attributes::SetNamedItem(const char * a_name, const char * a_value)
{
  Attribute ** attribute = &m_attributes;
  while(*attribute)
  {
    if(strcmp((*attribute)->GetName(), a_name) == 0)
    {
      if((*attribute)->m_value)
      {
        m_document->Free((void *) (*attribute)->m_value);
        (*attribute)->m_value = 0;
      }
      break;
    }
    attribute = &((*attribute)->m_next);
  }

  if(*attribute == NULL)
  {
    *attribute = ::new(m_document->Alloc(sizeof(Attribute))) Attribute();
  }

  (*attribute)->m_name = m_document->GetUniqueString(a_name);
  size_t size = MXMLUtf8Size(a_value);
  char * value = (char *) m_document->Alloc(size + 1);
  memcpy(value, a_value, size + 1);
  (*attribute)->m_value = value;
}


static int MXMLVsscanf(const char *str, const char *format, va_list ap)
{
  // can this be implemented in a more portable way?
  long p1 = va_arg(ap, long);
  long p2 = va_arg(ap, long);
  long p3 = va_arg(ap, long);
  long p4 = va_arg(ap, long);
  long p5 = va_arg(ap, long);
  long p6 = va_arg(ap, long);
  long p7 = va_arg(ap, long);
  return sscanf(str, format, p1, p2, p3, p4, p5, p6, p7);
}


int MXMLElement::Attributes::ScanNamedItem(const char * a_name, const char * a_format, ...)
{
  int rc = 0;
  va_list args;
  const char * value = GetNamedItem(a_name);
  va_start(args, a_format);
  rc = MXMLVsscanf(value, a_format, args);
  va_end(args);
  return rc;
}


void MXMLElement::Attributes::FormatNamedItem(const char * a_name, const char * a_format, ...)
{
  va_list argList;
  va_start(argList, a_format);

  // this is unsafe
  const size_t bufferSize = 1024;
  char buffer[bufferSize];

  vsprintf(buffer, a_format, argList);

  SetNamedItem(a_name, buffer);
  
  va_end(argList);
}


void MXMLElement::Attributes::RemoveNamedItem(const char * a_name)
{
  Attribute ** attribute = &m_attributes;
  while(*attribute)
  {
    if(strcmp((*attribute)->GetName(), a_name) == 0)
    {
      Attribute * remove = *attribute;
      *attribute = (*attribute)->m_next;
      if(remove->m_value)
      {
        m_document->Free((void *) remove->m_value);
      }
      m_document->Free(remove);
      break;
    }
    attribute = &((*attribute)->m_next);
  }
}


MXMLElement::Attributes::Attributes(MXMLDocument * a_document)
{
  m_document = a_document;
  m_attributes = NULL;
}


MXMLElement::Attributes::~Attributes()
{
  // dispose attributes
  while(m_attributes)
  {
    Attribute * attribute = m_attributes;
    m_attributes = attribute->m_next;
    if(attribute->m_value)
    {
      m_document->Free((void *) attribute->m_value);
    }
    m_document->Free(attribute);
  }
}


void MXMLElement::SetName(const char * a_name)
{
  m_name = m_document->GetUniqueString(a_name);
}


bool MXMLElement::IsName(const char * a_name) const
{
  if(m_name)
  {
    return (strcmp(a_name, m_name) == 0);
  }
  return false;
}


const char * MXMLElement::GetText() const
{
  const MXMLText * firstText = NULL;
  bool usedBuffer = false;

  for(const MXMLNode * node = GetFirstChild(); node; node = node->GetNextSibling())
  {
    if(node->IsType(MXMLNode::T_TEXT))
    {
      const MXMLText * text = static_cast<const MXMLText *>(node);
      if(firstText == NULL)
      {
        firstText = text;
      }
      else
      {
        if(!usedBuffer)
        {
          m_document->ResetBuffer();
          m_document->AppendBuffer(firstText->GetText());
          usedBuffer = true;
        }
        m_document->AppendBuffer(text->GetText());
      }
    }
  }

  if(!firstText)
  {
    return "";
  }
  else if(!usedBuffer)
  {
    return firstText->GetText();
  }
  return m_document->GetBuffer();
}


MXMLElement * MXMLElement::FindChild(const char * a_name)
{
  for(MXMLElement * element = GetFirstChildElement(); element; element = element->GetNextSiblingElement())
    if(element->IsName(a_name))
      return element;
  return NULL;
}


const MXMLElement * MXMLElement::FindChild(const char * a_name) const
{
  for(const MXMLElement * element = GetFirstChildElement(); element; element = element->GetNextSiblingElement())
    if(element->IsName(a_name))
      return element;
  return NULL;
}


MXMLElement * MXMLElement::GetFirstChildElement()
{
  for(MXMLNode * node = GetFirstChild(); node; node = node->GetNextSibling())
    if(node->IsType(MXMLNode::T_ELEMENT))
      return static_cast<MXMLElement *>(node);
  return NULL;
}


const MXMLElement * MXMLElement::GetFirstChildElement() const
{
  for(const MXMLNode * node = GetFirstChild(); node; node = node->GetNextSibling())
    if(node->IsType(MXMLNode::T_ELEMENT))
      return static_cast<const MXMLElement *>(node);
  return NULL;
}


MXMLElement * MXMLElement::GetNextSiblingElement()
{
  for(MXMLNode * node = GetNextSibling(); node; node = node->GetNextSibling())
    if(node->IsType(MXMLNode::T_ELEMENT))
      return static_cast<MXMLElement *>(node);
  return NULL;
}


const MXMLElement * MXMLElement::GetNextSiblingElement() const
{
  for(const MXMLNode * node = GetNextSibling(); node; node = node->GetNextSibling())
    if(node->IsType(MXMLNode::T_ELEMENT))
      return static_cast<const MXMLElement *>(node);
  return NULL;
}


bool MXMLElement::RemoveChild(MXMLNode * a_node)
{
  if(a_node->GetParent() == this)
  {
    if(a_node->m_prevSibling)
    {
      a_node->m_prevSibling->m_nextSibling = a_node->m_nextSibling;
    }
    if(a_node->m_nextSibling)
    {
      a_node->m_nextSibling->m_prevSibling = a_node->m_prevSibling;
    }
    if(m_firstChild == a_node)
    {
      m_firstChild = a_node->m_nextSibling;
    }
    if(m_lastChild == a_node)
    {
      m_lastChild= a_node->m_prevSibling;
    }

    a_node->m_nextSibling = NULL;
    a_node->m_prevSibling = NULL;
    a_node->m_parent = NULL;

    return true;
  }
  return false;
}


bool MXMLElement::InsertAfter(MXMLNode * a_cursor, MXMLNode * a_node)
{
  if(a_cursor->GetParent() != this || a_cursor == a_node)
  {
    return false;
  }

  assert(!a_node->GetParent());

  a_node->m_prevSibling = a_cursor;
  a_node->m_nextSibling = a_cursor->m_nextSibling;
  if(a_node->m_nextSibling)
    a_node->m_nextSibling->m_prevSibling = a_node;
  a_cursor->m_nextSibling = a_node;
  if(m_lastChild == a_cursor)
    m_lastChild = a_node;
  a_node->m_parent = this;

  return true;
}


bool MXMLElement::InsertBefore(MXMLNode * a_cursor, MXMLNode * a_node)
{
  if(a_cursor->GetParent() != this || a_cursor == a_node)
  {
    return false;
  }

  assert(!a_node->GetParent());

  a_node->m_nextSibling = a_cursor;
  a_node->m_prevSibling = a_cursor->m_prevSibling;
  if(a_node->m_prevSibling)
    a_node->m_prevSibling->m_nextSibling = a_node;
  a_cursor->m_prevSibling = a_node;
  if(m_firstChild == a_cursor)
    m_firstChild = a_node;
  a_node->m_parent = this;

  return true;
}


void MXMLElement::InsertLast(MXMLNode * a_node)
{
  if(a_node->GetParent())
  {
    a_node->GetParent()->RemoveChild(a_node);
  }

  a_node->m_prevSibling = m_lastChild;
  a_node->m_nextSibling = NULL;
  if(a_node->m_prevSibling)
  {
    a_node->m_prevSibling->m_nextSibling = a_node;
  }
  else
  {
    m_firstChild = a_node;
  }
  m_lastChild = a_node;

  a_node->m_parent = this;
}


MXMLDeclaration * MXMLElement::AppendDeclaration(const char * a_declaration, size_t a_size)
{
  MXMLDeclaration * declaration = m_document->CreateDeclaration();
  if(a_declaration)
  {
    declaration->SetText(a_declaration, a_size);
  }
  InsertLast(declaration);
  return declaration;
}


MXMLComment * MXMLElement::AppendComment(const char * a_comment, size_t a_size)
{
  MXMLComment * comment = m_document->CreateComment();
  if(a_comment)
  {
    comment->SetText(a_comment, a_size);
  }
  InsertLast(comment);
  return comment;
}


MXMLText * MXMLElement::AppendText(const char * a_text, size_t a_size)
{
  MXMLText * text = m_document->CreateText();
  if(a_text)
  {
    text->SetText(a_text, a_size);
  }
  InsertLast(text);
  return text;
}


MXMLCDATA * MXMLElement::AppendCDATA(const char * a_cdata, size_t a_size)
{
  MXMLCDATA * cdata = m_document->CreateCDATA();
  if(a_cdata)
  {
    cdata->SetText(a_cdata, a_size);
  }
  InsertLast(cdata);
  return cdata;
}


MXMLElement * MXMLElement::AppendElement(const char * a_name)
{
  MXMLElement * element = m_document->CreateElement();
  if(a_name)
  {
    element->SetName(a_name);
  }
  InsertLast(element);
  return element;
}


void MXMLElement::Format(int a_tabs)
{
  // first step is to trim all whitespace text
  for(MXMLNode * child = GetFirstChild(); child;)
  {
    MXMLNode * nextChild = child->GetNextSibling();
    if(child->IsType(MXMLNode::T_TEXT))
    {
      MXMLText * text = static_cast<MXMLText *>(child);
      const char * data = text->GetText();
      while(*data)
      {
        if(!(*data == ' ' || *data == '\r' || *data == '\n' || *data == '\t'))
        {
          break;
        }
        ++data;
      }

      if(*data)
      {
        // find the last valid character
        const char * end = data;
        const char * start = data;
        while(*data)
        {
          if(!(*data == ' ' || *data == '\r' || *data == '\n' || *data == '\t'))
          {
            end = data;
          }
          ++data;
        }
        ++end;
        if(*end || start != text->GetText())
        {
          text->SetText(start, end - start);
        }
      }
      else
      {
        child->Dispose();
      }
    }
    child = nextChild;
  }

  // add formatting
  if(GetFirstChild())
  {
    m_document->ResetBuffer();
    const char * cr = "\r\n";
    m_document->AppendBuffer(cr, strlen(cr));
    for(int i = 1; i < a_tabs; ++i)
    {
      m_document->AppendBuffer("  ", 2);
    }

    // we dont format if there is a single text node
    if(GetFirstChild() != GetLastChild() || !(GetFirstChild()->IsType(MXMLNode::T_TEXT) && !GetFirstChild()->IsType(MXMLNode::T_CDATA)))
    {
      if(a_tabs > 0)
      {
        MXMLText * closingText = m_document->CreateText();
        closingText->SetText(m_document->GetBuffer());
        m_document->AppendBuffer("  ", 2);
  
        // format our children
        for(MXMLNode * child = GetFirstChild(); child;)
        {
          MXMLNode * nextChild = child->GetNextSibling();
  
          // append CR + tabs before the child
          MXMLText * text = m_document->CreateText();
          text->SetText(m_document->GetBuffer());
          InsertBefore(child, text);
          child = nextChild;
        }
  
        InsertLast(closingText);
      }
      else
      {
        // format our children
        for(MXMLNode * child = GetFirstChild(); child;)
        {
          MXMLNode * nextChild = child->GetNextSibling();
  
          // append CR after the child
          MXMLText * text = m_document->CreateText();
          text->SetText(m_document->GetBuffer());
          InsertAfter(child, text);
          child = nextChild;
        }
      }

      // format children
      for(MXMLElement * element = GetFirstChildElement(); element; element = element->GetNextSiblingElement())
      {
        element->Format(a_tabs + 1);
      }
    }
  }
}


bool MXMLElement::Write(MXMLStream &a_stream) const
{
  int result = 0;
  size_t size = 0;
  const char * openA = "<";
  const char * openB = "</";
  const char * closeA = "/>";
  const char * closeB = ">";
  const char * space = " ";
  const char * assign = "=\"";
  const char * closeAttr = "\"";
  const char * name = GetName();

  size = strlen(openA);
  result |= (a_stream.Write(openA, size) != size);

  size = MXMLUtf8Size(name);
  result |= (a_stream.Write(name, size) != size);

  // write attributes
  const Attributes::Attribute * attribute = GetAttributes().GetFirstAttribute();
  while(attribute)
  {
    size = strlen(space);
    result |= (a_stream.Write(space, size) != size);

    size = MXMLUtf8Size(attribute->GetName());
    result |= (a_stream.Write(attribute->GetName(), size) != size);

    size = strlen(assign);
    result |= (a_stream.Write(assign, size) != size);

    size = MXMLUtf8Size(attribute->GetValue());
    result |= (a_stream.Write(attribute->GetValue(), size) != size);

    size = strlen(closeAttr);
    result |= (a_stream.Write(closeAttr, size) != size);

    attribute = attribute->GetNextAttribute();
  }

  // if no children, write shorthand element
  if(GetFirstChild() == NULL)
  {
    size = strlen(closeA);
    result |= (a_stream.Write(closeA, size) != size);
  }
  else
  {
    size = strlen(closeB);
    result |= (a_stream.Write(closeB, size) != size);

    // write children
    for(const MXMLNode * child = GetFirstChild(); child != NULL; child = child->GetNextSibling())
    {
      if(!child->Write(a_stream))
      {
        result |= 1;
        break;
      }
    }

    // write name close
    size = strlen(openB);
    result |= (a_stream.Write(openB, size) != size);

    size = MXMLUtf8Size(name);
    result |= (a_stream.Write(name, size) != size);

    size = strlen(closeB);
    result |= (a_stream.Write(closeB, size) != size);
  }

  return (result == 0);
}


MXMLElement::MXMLElement(MXMLDocument * a_document) : 
  MXMLNode(a_document),
  m_attributes(a_document)
{
  m_type |= T_ELEMENT;
  m_name = NULL;
  m_firstChild = NULL;
  m_lastChild = NULL;
}


MXMLElement::~MXMLElement()
{
  while(GetFirstChild())
  {
    GetFirstChild()->Dispose();
  }
}


MXMLDocument::MXMLDocument()
{
  m_uniques = NULL;
  m_sizeUniques = 233;

  m_buffer = NULL;
  m_bufferSize = 0;
  m_maxBufferSize = 0;

  m_Alloc = malloc;
  m_Free = free;

  // create a document element
  m_document = CreateElement();
}


MXMLDocument::MXMLDocument(MXMLAlloc a_Alloc, MXMLFree a_Free)
{
  m_uniques = NULL;
  m_sizeUniques = 233;

  m_buffer = NULL;
  m_bufferSize = 0;
  m_maxBufferSize = 0;

  m_Alloc = a_Alloc;
  m_Free = a_Free;

  // create a document element
  m_document = CreateElement();
}


MXMLDocument::~MXMLDocument()
{
  // dispose root
  m_document->Dispose();

  // clear uniques
  if(m_uniques)
  {
    UniqueNode * node;
    for(int i = 0; i < m_sizeUniques; ++i)
    {
      while(m_uniques[i])
      {
        node = m_uniques[i];
        m_uniques[i] = m_uniques[i]->m_next;
        Free((void *) node->m_string);
        Free(node);
      }
    }
    Free(m_uniques);
  }

  // free buffer
  if(m_buffer)
  {
    Free(m_buffer);
  }
}


MXMLDeclaration * MXMLDocument::CreateDeclaration()
{
  return ::new(Alloc(sizeof(MXMLDeclaration))) MXMLDeclaration(this);
}


MXMLComment * MXMLDocument::CreateComment()
{
  return ::new(Alloc(sizeof(MXMLComment))) MXMLComment(this);
}


MXMLText * MXMLDocument::CreateText()
{
  return ::new(Alloc(sizeof(MXMLText))) MXMLText(this);
}


MXMLCDATA * MXMLDocument::CreateCDATA()
{
  return ::new(Alloc(sizeof(MXMLCDATA))) MXMLCDATA(this);
}


MXMLElement * MXMLDocument::CreateElement()
{
  return ::new(Alloc(sizeof(MXMLElement))) MXMLElement(this);
}


void MXMLDocument::ResetBuffer()
{
  m_bufferSize = 1;
  if(m_buffer == NULL)
  {
    m_maxBufferSize = 1024;
    m_buffer = (char *) Alloc(sizeof(char) * m_maxBufferSize);
  }
  m_buffer[0] = '\0';
}


void MXMLDocument::AppendBuffer(const char * a_text, size_t a_size)
{
  size_t size = (a_size) ? a_size : MXMLUtf8Size(a_text);
  size_t newSize = m_bufferSize + size;
  if(newSize >= m_maxBufferSize)
  {
    --newSize;

    newSize |= newSize >> 16;
    newSize |= newSize >> 8;
    newSize |= newSize >> 4;
    newSize |= newSize >> 2;
    newSize |= newSize >> 1;

    ++newSize;

    m_maxBufferSize = newSize;
    char * buffer = (char *) Alloc(sizeof(char) * m_maxBufferSize);
    memcpy(buffer, m_buffer, m_bufferSize);
    Free(m_buffer);
    m_buffer = buffer;
  }

  memcpy(m_buffer + m_bufferSize - 1, a_text, size);
  m_bufferSize += size;
  m_buffer[m_bufferSize - 1] = '\0';
}


const char * MXMLDocument::GetBuffer() const
{
  if(m_buffer)
  {
    return m_buffer;
  }
  return "";
}


struct MXMLUserData
{
  MXMLDocument * m_document;
  MXMLElement * m_root;
  MXMLElement * m_element;
};


static void MXMLStartElement(void * a_userData, const char * a_name, const char ** a_attributes)
{
  MXMLUserData * userData = (MXMLUserData *) a_userData;

  MXMLElement * element = userData->m_element->AppendElement(a_name);

  for(int i = 0; a_attributes[i]; i += 2)
  {
    element->GetAttributes().SetNamedItem(a_attributes[i], a_attributes[i + 1]);
  }

  userData->m_element = element;
}


static void MXMLEndElement(void * a_userData, const char * a_name)
{
  a_name;
  MXMLUserData * userData = (MXMLUserData *) a_userData;
  if(userData->m_element->GetParent())
  {
    userData->m_element = userData->m_element->GetParent();
  }
}


static void MXMLCharacterData(void * a_userData, const char * a_data, size_t a_length, MXMLCharacterDataType a_type)
{
  MXMLUserData * userData = (MXMLUserData *) a_userData;
  switch(a_type)
  {
    case MXML_CDT_DECLARATION :
      userData->m_element->AppendDeclaration(a_data, a_length);
      break;
    case MXML_CDT_COMMENT :
      userData->m_element->AppendComment(a_data, a_length);
      break;
    case MXML_CDT_TEXT :
      userData->m_element->AppendText(a_data, a_length);
      break;
    case MXML_CDT_CDATA :
      userData->m_element->AppendCDATA(a_data, a_length);
      break;
    default:
      break;
  }
}


bool MXMLDocument::Load(MXMLStream &a_stream, size_t a_bufferSize)
{
  unsigned char * buffer = (unsigned char *) Alloc(a_bufferSize);
  MXMLParser * parser = MXMLParserCreate(m_Alloc, m_Free);
  int done;
  
  MXMLUserData userData;
  userData.m_document = this;
  userData.m_root = CreateElement();
  userData.m_element = userData.m_root;

  parser->m_userData = &userData;
  parser->m_StartElementCallback = MXMLStartElement;
  parser->m_EndElementCallback = MXMLEndElement;
  parser->m_CharacterDataCallback = MXMLCharacterData;

  do 
  {
    size_t len = a_stream.Read(buffer, a_bufferSize);
    done = len < a_bufferSize;
    if (MXMLParserParse(parser, buffer, len, done) == MXML_RESULT_ERROR) 
    {
      userData.m_root->Dispose();

      ResetBuffer();
      AppendBuffer(MXMLParserGetErrorString(parser));
      MXMLParserDestroy(parser);
      Free(buffer);
      return false;
    }
  } while (!done);

  m_document->Dispose();
  m_document = userData.m_root;

  MXMLParserDestroy(parser);
  Free(buffer);

  return true;
}


bool MXMLDocument::Save(MXMLStream &a_stream) const
{
  for(MXMLNode * node = m_document->GetFirstChild(); node; node = node->GetNextSibling())
  {
    node->Write(a_stream);
  }
  return true;
}


const char * MXMLDocument::GetUniqueString(const char * a_string)
{
  // init the table if not yet initialized
  if(m_uniques == NULL)
  {
    m_uniques = (UniqueNode **) Alloc(sizeof(UniqueNode *) * m_sizeUniques);
    for(int i = 0; i < m_sizeUniques; ++i)
    {
      m_uniques[i] = NULL;
    }
  }

  // get hash slot
  unsigned int slot = MXMLUtf8Hash(a_string) % m_sizeUniques;

  // find insert string
  UniqueNode ** node = &m_uniques[slot];
  while(*node)
  {
    int compare = strcmp((*node)->m_string, a_string);
    if(compare == 0)
    {
      return (*node)->m_string;
    }
    else if(compare > 0)
    {
      break;
    }
    node = &((*node)->m_next);
  }

  UniqueNode * newNode = (UniqueNode *) Alloc(sizeof(UniqueNode));
  size_t size = MXMLUtf8Size(a_string);
  char * string = (char * ) Alloc(size + 1);
  memcpy(string, a_string, size + 1);
  newNode->m_string = string;
  newNode->m_next = *node;
  *node = newNode;

  return newNode->m_string;
}

